from typing import Any, Callable, Dict, Union
import toml
import yaml

DEFAULT_PYPROJECT_TOML_FILENAME: str = 'pyproject.toml'
DEFAULT_CITATION_CFF_FILENAME: str = 'CITATION.cff'
DEFAULT_DIR: str = '.'

DEFAULT_PYPROJECT_TOML_FILE_PATH: str = DEFAULT_DIR + \
    DEFAULT_PYPROJECT_TOML_FILENAME
DEFAULT_CITATION_CFF_FILE_PATH: str = DEFAULT_DIR + DEFAULT_CITATION_CFF_FILENAME

TOMLObject = Dict[str, Any]
CFFObject = Any

TransformCFFObjectWithTOMLObjectFunction = Callable[[
    CFFObject, TOMLObject], CFFObject]
TransformTOMLObjectWithCFFObjectFunction = Callable[[
    TOMLObject, CFFObject], TOMLObject]


class LoadTOMLFileException(Exception):
    pass


class SaveTOMLFileException(Exception):
    pass


class LoadCFFFileException(Exception):
    pass


class SaveCFFFileException(Exception):
    pass


def load_toml_object(toml_file_path: str) -> TOMLObject:
    try:
        with open(toml_file_path, 'r') as toml_file:
            toml_object: TOMLObject = toml.load(toml_file)
    except:
        raise LoadTOMLFileException(
            f"Cannot load this TOML file: {toml_file_path}")
    return toml_object


def save_toml_object(toml_object: TOMLObject, toml_file_path: str) -> None:
    try:
        with open(toml_file_path, 'w') as toml_file:
            toml.dump(toml_object, toml_file)
    except:
        raise SaveTOMLFileException(
            f"Cannot save this TOML file: {toml_file_path}")


def load_cff_object(cff_file_path: str) -> CFFObject:
    try:
        with open(cff_file_path, 'r') as cff_file:
            cff_object: CFFObject = yaml.safe_load(cff_file)
    except:
        raise LoadCFFFileException(
            f"Cannot load this CFF file: {cff_file_path}")
    return cff_object


def save_cff_object(cff_object: CFFObject, cff_file_path: str) -> None:
    try:
        with open(cff_file_path, 'w') as cff_file:
            yaml.safe_dump(cff_object, cff_file)
    except:
        raise SaveCFFFileException(
            f"Cannot load this CFF file: {cff_file_path}")


def update_toml_with_cff(toml_file_path: str, cff_file_path: str, transform_toml_object_func: TransformTOMLObjectWithCFFObjectFunction) -> TOMLObject:

    toml_object: TOMLObject = load_toml_object(toml_file_path=toml_file_path)
    cff_object: CFFObject = load_cff_object(cff_file_path=cff_file_path)

    toml_object: TOMLObject = transform_toml_object_func(
        toml_object, cff_object)

    save_toml_object(toml_object=toml_object,
                     toml_file_path=toml_file_path)

    return toml_object


def update_cff_with_toml(toml_file_path: str, cff_file_path: str, transform_cff_object_func: TransformCFFObjectWithTOMLObjectFunction) -> CFFObject:

    toml_object: TOMLObject = load_toml_object(toml_file_path=toml_file_path)
    cff_object: CFFObject = load_cff_object(cff_file_path=cff_file_path)

    cff_object: CFFObject = transform_cff_object_func(cff_object, toml_object)

    save_cff_object(cff_object=cff_object,
                    cff_file_path=cff_file_path)

    return cff_object


def update_pyproject_toml_with_citation_cff(pyproject_toml_file_path=DEFAULT_PYPROJECT_TOML_FILE_PATH, citation_cff_file_path=DEFAULT_CITATION_CFF_FILE_PATH) -> TOMLObject:
    def transformer(pyproject_toml_object: TOMLObject, citation_cff_object: CFFObject) -> TOMLObject:
        pyproject_toml_object['project']['name'] = citation_cff_object['title']
        pyproject_toml_object['project']['version'] = citation_cff_object['version']
        pyproject_toml_object['project']['description'] = citation_cff_object['abstract']
        pyproject_toml_object['project']['license'] = citation_cff_object['license']
        pyproject_toml_object['project']['urls']['Source'] = citation_cff_object['repository-code']

        return pyproject_toml_object
    return update_toml_with_cff(toml_file_path=pyproject_toml_file_path, cff_file_path=citation_cff_file_path, transform_toml_object_func=transformer)


def update_citation_cff_with_pyproject_toml(pyproject_toml_file_path=DEFAULT_PYPROJECT_TOML_FILE_PATH, citation_cff_file_path=DEFAULT_CITATION_CFF_FILE_PATH) -> CFFObject:
    def transformer(citation_cff_object: CFFObject, pyproject_toml_object: TOMLObject) -> CFFObject:
        citation_cff_object['title'] = pyproject_toml_object['project']['name']
        citation_cff_object['version'] = pyproject_toml_object['project']['version']
        citation_cff_object['abstract'] = pyproject_toml_object['project']['description']
        citation_cff_object['license'] = pyproject_toml_object['project']['license']
        citation_cff_object['repository-code'] = pyproject_toml_object['project']['urls']['Source']

        return citation_cff_object
    return update_cff_with_toml(toml_file_path=pyproject_toml_file_path, cff_file_path=citation_cff_file_path, transform_cff_object_func=transformer)



import pytest
import os
from src.cff2toml.cff2toml import load_cff_object, load_toml_object, CFFObject, TOMLObject, TransformCFFObjectWithTOMLObjectFunction, TransformTOMLObjectWithCFFObjectFunction, update_cff_with_toml, update_citation_cff_with_pyproject_toml, update_pyproject_toml_with_citation_cff, update_toml_with_cff

import tempfile
import shutil
import os


class TempCopiedFile:

    def __init__(self, source_file_path: str):
        self.source_file_path = source_file_path

    def __enter__(self):
        self.tmp = tempfile.NamedTemporaryFile(delete=False)
        shutil.copy2(src=self.source_file_path, dst=self.tmp.name)
        return self

    def __exit__(self, *args):
        os.remove(path=self.tmp.name)

    @property
    def file_path(self) -> str:
        if self.tmp.closed:
            return ''
        return self.tmp.name


@pytest.fixture
def dummy_directory_file_path():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dummy_files')


@pytest.fixture
def dummy_citation_cff_file_path(dummy_directory_file_path):
    return os.path.join(dummy_directory_file_path, 'dummy_CITATION.cff')


@pytest.fixture
def dummy_pyproject_toml_file_path(dummy_directory_file_path):
    return os.path.join(dummy_directory_file_path, 'dummy_pyproject.toml')


def test_load_cff_object(dummy_citation_cff_file_path):
    with TempCopiedFile(source_file_path=dummy_citation_cff_file_path) as tmp_dummy_citation_cff_file:
        cff_object: CFFObject = load_cff_object(
            cff_file_path=tmp_dummy_citation_cff_file.file_path)
        assert cff_object['version'] == '0.0.2'
        assert cff_object['title'] == 'somecooltool'
        assert cff_object['license'] == 'Apache-2.0'
        assert cff_object['abstract'] == 'A module that does something cool.'
        assert cff_object['repository-code'] == 'https://github.com/willynilly/somewherecool'


def test_load_toml_object(dummy_pyproject_toml_file_path):
    with TempCopiedFile(source_file_path=dummy_pyproject_toml_file_path) as tmp_dummy_pyproject_toml_file:
        toml_object: TOMLObject = load_toml_object(
            toml_file_path=tmp_dummy_pyproject_toml_file.file_path)
        assert toml_object['project']['version'] == '0.0.1'
        assert toml_object['project']['name'] == 'someuncooltool'
        assert toml_object['project']['description'] == 'A module that does something uncool.'
        assert toml_object['project']['license'] == 'Apache-1.0'
        assert toml_object['project']['urls']['Source'] == 'https://github.com/willnilly/somewhereuncool'


def test_update_toml_with_cff(dummy_citation_cff_file_path, dummy_pyproject_toml_file_path):
    with TempCopiedFile(source_file_path=dummy_citation_cff_file_path) as tmp_dummy_citation_cff_file:
        with TempCopiedFile(source_file_path=dummy_pyproject_toml_file_path) as tmp_dummy_pyproject_toml_file:

            toml_object: TOMLObject = load_toml_object(
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path)
            assert toml_object['project']['version'] == '0.0.1'
            assert toml_object['project']['name'] == 'someuncooltool'
            assert toml_object['project']['description'] == 'A module that does something uncool.'
            assert toml_object['project']['license'] == 'Apache-1.0'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willnilly/somewhereuncool'

            def transformer(toml_object: TOMLObject, cff_object: CFFObject) -> TOMLObject:
                toml_object['project']['version'] = '5.0.1'
                toml_object['project']['name'] = 'cows'
                toml_object['project']['description'] = 'A module that does something else.'
                toml_object['project']['license'] = 'MIT'
                toml_object['project']['urls']['Source'] = 'https://github.com/willynilly/somewhereelse'
                return toml_object

            toml_object = update_toml_with_cff(
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path,
                cff_file_path=tmp_dummy_citation_cff_file.file_path,
                transform_toml_object_func=transformer)

            assert toml_object['project']['version'] == '5.0.1'
            assert toml_object['project']['name'] == 'cows'
            assert toml_object['project']['description'] == 'A module that does something else.'
            assert toml_object['project']['license'] == 'MIT'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willynilly/somewhereelse'

            toml_object = load_toml_object(
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path)
            assert toml_object['project']['version'] == '5.0.1'
            assert toml_object['project']['name'] == 'cows'
            assert toml_object['project']['description'] == 'A module that does something else.'
            assert toml_object['project']['license'] == 'MIT'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willynilly/somewhereelse'


def test_update_cff_with_toml(dummy_citation_cff_file_path, dummy_pyproject_toml_file_path):
    with TempCopiedFile(source_file_path=dummy_citation_cff_file_path) as tmp_dummy_citation_cff_file:
        with TempCopiedFile(source_file_path=dummy_pyproject_toml_file_path) as tmp_dummy_pyproject_toml_file:

            cff_object: CFFObject = load_cff_object(
                cff_file_path=tmp_dummy_citation_cff_file.file_path)
            assert cff_object['version'] == '0.0.2'
            assert cff_object['title'] == 'somecooltool'
            assert cff_object['license'] == 'Apache-2.0'
            assert cff_object['abstract'] == 'A module that does something cool.'
            assert cff_object['repository-code'] == 'https://github.com/willynilly/somewherecool'

            def transformer(cff_object: CFFObject, toml_object: TOMLObject) -> CFFObject:
                cff_object['version'] = '4.0.0'
                cff_object['title'] = 'sheep'
                cff_object['license'] = 'MIT'
                cff_object['abstract'] = 'A module that does something else.'
                cff_object['repository-code'] = 'https://github.com/willynilly/somewhereelse'

                return cff_object

            cff_object = update_cff_with_toml(
                cff_file_path=tmp_dummy_citation_cff_file.file_path,
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path,
                transform_cff_object_func=transformer)

            assert cff_object['version'] == '4.0.0'
            assert cff_object['title'] == 'sheep'
            assert cff_object['license'] == 'MIT'
            assert cff_object['abstract'] == 'A module that does something else.'
            assert cff_object['repository-code'] == 'https://github.com/willynilly/somewhereelse'

            cff_object = load_cff_object(
                cff_file_path=tmp_dummy_citation_cff_file.file_path)
            assert cff_object['version'] == '4.0.0'
            assert cff_object['title'] == 'sheep'
            assert cff_object['license'] == 'MIT'
            assert cff_object['abstract'] == 'A module that does something else.'
            assert cff_object['repository-code'] == 'https://github.com/willynilly/somewhereelse'


def test_update_pyproject_toml_with_citation_cff(dummy_citation_cff_file_path, dummy_pyproject_toml_file_path):
    with TempCopiedFile(source_file_path=dummy_citation_cff_file_path) as tmp_dummy_citation_cff_file:
        with TempCopiedFile(source_file_path=dummy_pyproject_toml_file_path) as tmp_dummy_pyproject_toml_file:

            toml_object: TOMLObject = load_toml_object(
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path)
            assert toml_object['project']['version'] == '0.0.1'
            assert toml_object['project']['name'] == 'someuncooltool'
            assert toml_object['project']['description'] == 'A module that does something uncool.'
            assert toml_object['project']['license'] == 'Apache-1.0'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willnilly/somewhereuncool'

            toml_object = update_pyproject_toml_with_citation_cff(
                pyproject_toml_file_path=tmp_dummy_pyproject_toml_file.file_path,
                citation_cff_file_path=tmp_dummy_citation_cff_file.file_path)

            assert toml_object['project']['version'] == '0.0.2'
            assert toml_object['project']['name'] == 'somecooltool'
            assert toml_object['project']['description'] == 'A module that does something cool.'
            assert toml_object['project']['license'] == 'Apache-2.0'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willynilly/somewherecool'

            toml_object = load_toml_object(
                toml_file_path=tmp_dummy_pyproject_toml_file.file_path)
            assert toml_object['project']['version'] == '0.0.2'
            assert toml_object['project']['name'] == 'somecooltool'
            assert toml_object['project']['description'] == 'A module that does something cool.'
            assert toml_object['project']['license'] == 'Apache-2.0'
            assert toml_object['project']['urls']['Source'] == 'https://github.com/willynilly/somewherecool'


def test_update_citation_cff_with_pyproject_toml(dummy_citation_cff_file_path, dummy_pyproject_toml_file_path):
    with TempCopiedFile(source_file_path=dummy_citation_cff_file_path) as tmp_dummy_citation_cff_file:
        with TempCopiedFile(source_file_path=dummy_pyproject_toml_file_path) as tmp_dummy_pyproject_toml_file:

            cff_object: CFFObject = load_cff_object(
                cff_file_path=tmp_dummy_citation_cff_file.file_path)
            assert cff_object['version'] == '0.0.2'
            assert cff_object['title'] == 'somecooltool'
            assert cff_object['license'] == 'Apache-2.0'
            assert cff_object['abstract'] == 'A module that does something cool.'
            assert cff_object['repository-code'] == 'https://github.com/willynilly/somewherecool'

            cff_object = update_citation_cff_with_pyproject_toml(
                citation_cff_file_path=tmp_dummy_citation_cff_file.file_path,
                pyproject_toml_file_path=tmp_dummy_pyproject_toml_file.file_path)

            assert cff_object['version'] == '0.0.1'
            assert cff_object['title'] == 'someuncooltool'
            assert cff_object['license'] == 'Apache-1.0'
            assert cff_object['abstract'] == 'A module that does something uncool.'
            assert cff_object['repository-code'] == 'https://github.com/willnilly/somewhereuncool'

            cff_object = load_cff_object(
                cff_file_path=tmp_dummy_citation_cff_file.file_path)
            assert cff_object['version'] == '0.0.1'
            assert cff_object['title'] == 'someuncooltool'
            assert cff_object['license'] == 'Apache-1.0'
            assert cff_object['abstract'] == 'A module that does something uncool.'
            assert cff_object['repository-code'] == 'https://github.com/willnilly/somewhereuncool'
